@@ -1,5 +1,5 @@
 # Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
-# Copyright 2011 OpenStack, LLC
+# Copyright (c) 2011 OpenStack Foundation
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -116,6 +116,34 @@ def _stub_restore():
     return {'volume_id': '712f4980-5ac1-41e5-9383-390aa7c9f58b'}
 
 
+def _stub_qos_full(id, base_uri, tenant_id, name=None, specs=None):
+    if not name:
+        name = 'fake-name'
+    if not specs:
+        specs = {}
+
+    return {
+        'qos_specs': {
+            'id': id,
+            'name': name,
+            'consumer': 'back-end',
+            'specs': specs,
+        },
+        'links': {
+            'href': _bookmark_href(base_uri, tenant_id, id),
+            'rel': 'bookmark'
+        }
+    }
+
+
+def _stub_qos_associates(id, name):
+    return {
+        'assoications_type': 'volume_type',
+        'name': name,
+        'id': id,
+    }
+
+
 def _stub_transfer_full(id, base_uri, tenant_id):
     return {
         'id': id,
@@ -240,12 +268,14 @@ class FakeHTTPClient(base_client.HTTPClient):
     def post_snapshots_1234_action(self, body, **kw):
         _body = None
         resp = 202
-        assert len(body.keys()) == 1
-        action = body.keys()[0]
+        assert len(list(body)) == 1
+        action = list(body)[0]
         if action == 'os-reset_status':
             assert 'status' in body['os-reset_status']
+        elif action == 'os-update_snapshot_status':
+            assert 'status' in body['os-update_snapshot_status']
         else:
-            raise AssertionError('Unexpected action: %s" % action')
+            raise AssertionError("Unexpected action: %s" % action)
         return (resp, {}, _body)
 
     #
@@ -267,22 +297,32 @@ class FakeHTTPClient(base_client.HTTPClient):
     # at the very least it's not complete
     def get_volumes_detail(self, **kw):
         return (200, {}, {"volumes": [
-            {'id': 1234,
+            {'id': kw.get('id', 1234),
              'name': 'sample-volume',
              'attachments': [{'server_id': 1234}]},
         ]})
 
     def get_volumes_1234(self, **kw):
-        r = {'volume': self.get_volumes_detail()[2]['volumes'][0]}
+        r = {'volume': self.get_volumes_detail(id=1234)[2]['volumes'][0]}
+        return (200, {}, r)
+
+    def get_volumes_5678(self, **kw):
+        r = {'volume': self.get_volumes_detail(id=5678)[2]['volumes'][0]}
+        return (200, {}, r)
+
+    def get_volumes_1234_encryption(self, **kw):
+        r = {'encryption_key_id': 'id'}
         return (200, {}, r)
 
     def post_volumes_1234_action(self, body, **kw):
         _body = None
         resp = 202
-        assert len(list(body.keys())) == 1
-        action = list(body.keys())[0]
+        assert len(list(body)) == 1
+        action = list(body)[0]
         if action == 'os-attach':
-            assert list(body[action].keys()) == ['instance_uuid', 'mountpoint']
+            assert sorted(list(body[action])) == ['instance_uuid',
+                                                  'mode',
+                                                  'mountpoint']
         elif action == 'os-detach':
             assert body[action] is None
         elif action == 'os-reserve':
@@ -290,10 +330,10 @@ class FakeHTTPClient(base_client.HTTPClient):
         elif action == 'os-unreserve':
             assert body[action] is None
         elif action == 'os-initialize_connection':
-            assert list(body[action].keys()) == ['connector']
+            assert list(body[action]) == ['connector']
             return (202, {}, {'connection_info': 'foos'})
         elif action == 'os-terminate_connection':
-            assert list(body[action].keys()) == ['connector']
+            assert list(body[action]) == ['connector']
         elif action == 'os-begin_detaching':
             assert body[action] is None
         elif action == 'os-roll_detaching':
@@ -301,7 +341,12 @@ class FakeHTTPClient(base_client.HTTPClient):
         elif action == 'os-reset_status':
             assert 'status' in body[action]
         elif action == 'os-extend':
-            assert body[action].keys() == ['new_size']
+            assert list(body[action]) == ['new_size']
+        elif action == 'os-migrate_volume':
+            assert 'host' in body[action]
+            assert 'force_host_copy' in body[action]
+        elif action == 'os-update_readonly_flag':
+            assert list(body[action]) == ['readonly']
         else:
             raise AssertionError("Unexpected action: %s" % action)
         return (resp, {}, _body)
@@ -312,6 +357,9 @@ class FakeHTTPClient(base_client.HTTPClient):
     def delete_volumes_1234(self, **kw):
         return (202, {}, None)
 
+    def delete_volumes_5678(self, **kw):
+        return (202, {}, None)
+
     #
     # Quotas
     #
@@ -333,7 +381,7 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'gigabytes': 1}})
 
     def put_os_quota_sets_test(self, body, **kw):
-        assert list(body.keys()) == ['quota_set']
+        assert list(body) == ['quota_set']
         fakes.assert_has_keys(body['quota_set'],
                               required=['tenant_id'])
         return (200, {}, {'quota_set': {
@@ -356,7 +404,7 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'gigabytes': 1}})
 
     def put_os_quota_class_sets_test(self, body, **kw):
-        assert list(body.keys()) == ['quota_class_set']
+        assert list(body) == ['quota_class_set']
         fakes.assert_has_keys(body['quota_class_set'],
                               required=['class_name'])
         return (200, {}, {'quota_class_set': {
@@ -383,13 +431,18 @@ class FakeHTTPClient(base_client.HTTPClient):
                           'name': 'test-type-1',
                           'extra_specs': {}}})
 
+    def get_types_2(self, **kw):
+        return (200, {}, {'volume_type': {'id': 2,
+                          'name': 'test-type-2',
+                          'extra_specs': {}}})
+
     def post_types(self, body, **kw):
         return (202, {}, {'volume_type': {'id': 3,
                           'name': 'test-type-3',
                           'extra_specs': {}}})
 
     def post_types_1_extra_specs(self, body, **kw):
-        assert list(body.keys()) == ['extra_specs']
+        assert list(body) == ['extra_specs']
         return (200, {}, {'extra_specs': {'k': 'v'}})
 
     def delete_types_1_extra_specs_k(self, **kw):
@@ -398,6 +451,23 @@ class FakeHTTPClient(base_client.HTTPClient):
     def delete_types_1(self, **kw):
         return (202, {}, None)
 
+    #
+    # VolumeEncryptionTypes
+    #
+    def get_types_1_encryption(self, **kw):
+        return (200, {}, {'id': 1, 'volume_type_id': 1, 'provider': 'test',
+                          'cipher': 'test', 'key_size': 1,
+                          'control_location': 'front'})
+
+    def get_types_2_encryption(self, **kw):
+        return (200, {}, {})
+
+    def post_types_2_encryption(self, body, **kw):
+        return (200, {}, {'encryption': {}})
+
+    def put_types_1_encryption_1(self, body, **kw):
+        return (200, {}, {})
+
     #
     # Set/Unset metadata
     #
@@ -474,6 +544,72 @@ class FakeHTTPClient(base_client.HTTPClient):
         return (200, {},
                 {'restore': _stub_restore()})
 
+    def post_backups_1234_restore(self, **kw):
+        return (200, {},
+                {'restore': _stub_restore()})
+
+    #
+    # QoSSpecs
+    #
+
+    def get_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C(self, **kw):
+        base_uri = 'http://localhost:8776'
+        tenant_id = '0fa851f6668144cf9cd8c8419c1646c1'
+        qos_id1 = '1B6B6A04-A927-4AEB-810B-B7BAAD49F57C'
+        return (200, {},
+                _stub_qos_full(qos_id1, base_uri, tenant_id))
+
+    def get_qos_specs(self, **kw):
+        base_uri = 'http://localhost:8776'
+        tenant_id = '0fa851f6668144cf9cd8c8419c1646c1'
+        qos_id1 = '1B6B6A04-A927-4AEB-810B-B7BAAD49F57C'
+        qos_id2 = '0FD8DD14-A396-4E55-9573-1FE59042E95B'
+        return (200, {},
+                {'qos_specs': [
+                    _stub_qos_full(qos_id1, base_uri, tenant_id, 'name-1'),
+                    _stub_qos_full(qos_id2, base_uri, tenant_id)]})
+
+    def post_qos_specs(self, **kw):
+        base_uri = 'http://localhost:8776'
+        tenant_id = '0fa851f6668144cf9cd8c8419c1646c1'
+        qos_id = '1B6B6A04-A927-4AEB-810B-B7BAAD49F57C'
+        qos_name = 'qos-name'
+        return (202, {},
+                _stub_qos_full(qos_id, base_uri, tenant_id, qos_name))
+
+    def put_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C(self, **kw):
+        return (202, {}, None)
+
+    def put_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C_delete_keys(
+            self, **kw):
+        return (202, {}, None)
+
+    def delete_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C(self, **kw):
+        return (202, {}, None)
+
+    def get_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C_associations(
+            self, **kw):
+        type_id1 = '4230B13A-7A37-4E84-B777-EFBA6FCEE4FF'
+        type_id2 = '4230B13A-AB37-4E84-B777-EFBA6FCEE4FF'
+        type_name1 = 'type1'
+        type_name2 = 'type2'
+        return (202, {},
+                {'qos_associations': [
+                    _stub_qos_associates(type_id1, type_name1),
+                    _stub_qos_associates(type_id2, type_name2)]})
+
+    def get_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C_associate(
+            self, **kw):
+        return (202, {}, None)
+
+    def get_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C_disassociate(
+            self, **kw):
+        return (202, {}, None)
+
+    def get_qos_specs_1B6B6A04_A927_4AEB_810B_B7BAAD49F57C_disassociate_all(
+            self, **kw):
+        return (202, {}, None)
+
     #
     # VolumeTransfers
     #
@@ -613,3 +749,18 @@ class FakeHTTPClient(base_client.HTTPClient):
                 },
                 ]
         })
+
+    def post_snapshots_1234_metadata(self, **kw):
+        return (200, {}, {"metadata": {"key1": "val1", "key2": "val2"}})
+
+    def delete_snapshots_1234_metadata_key1(self, **kw):
+        return (200, {}, None)
+
+    def delete_snapshots_1234_metadata_key2(self, **kw):
+        return (200, {}, None)
+
+    def put_volumes_1234_metadata(self, **kw):
+        return (200, {}, {"metadata": {"key1": "val1", "key2": "val2"}})
+
+    def put_snapshots_1234_metadata(self, **kw):
+        return (200, {}, {"metadata": {"key1": "val1", "key2": "val2"}})
