@@ -37,13 +37,14 @@ class Volume(base.Resource):
         """Update the name or description for this volume."""
         self.manager.update(self, **kwargs)
 
-    def attach(self, instance_uuid, mountpoint):
+    def attach(self, instance_uuid, mountpoint, mode='rw'):
         """Set attachment metadata.
 
         :param instance_uuid: uuid of the attaching instance.
         :param mountpoint: mountpoint on the attaching instance.
+        :param mode: the access mode.
         """
-        return self.manager.attach(self, instance_uuid, mountpoint)
+        return self.manager.attach(self, instance_uuid, mountpoint, mode)
 
     def detach(self):
         """Clear attachment metadata."""
@@ -121,6 +122,19 @@ class Volume(base.Resource):
 #        self.manager.migrate_volume_completion(self, old_volume,
 #                                               new_volume, error)
 
+    def update_all_metadata(self, metadata):
+        """Update all metadata of this volume."""
+        return self.manager.update_all_metadata(self, metadata)
+
+    def update_readonly_flag(self, volume, read_only):
+        """Update the read-only access mode flag of the specified volume.
+
+        :param volume: The UUID of the volume to update.
+        :param read_only: The value to indicate whether to update volume to
+            read-only access mode.
+        """
+        self.manager.update_readonly_flag(self, volume, read_only)
+
 
 class VolumeManager(base.ManagerWithFind):
     """Manage :class:`Volume` resources."""
@@ -229,18 +243,20 @@ class VolumeManager(base.ManagerWithFind):
         url = '/volumes/%s/action' % base.getid(volume)
         return self.api.client.post(url, body=body)
 
-    def attach(self, volume, instance_uuid, mountpoint):
+    def attach(self, volume, instance_uuid, mountpoint, mode='rw'):
         """Set attachment metadata.
 
         :param volume: The :class:`Volume` (or its ID)
                        you would like to attach.
         :param instance_uuid: uuid of the attaching instance.
         :param mountpoint: mountpoint on the attaching instance.
+        :param mode: the access mode.
         """
         return self._action('os-attach',
                             volume,
                             {'instance_uuid': instance_uuid,
-                             'mountpoint': mountpoint})
+                             'mountpoint': mountpoint,
+                             'mode': mode})
 
     def detach(self, volume):
         """Clear attachment metadata.
@@ -314,7 +330,7 @@ class VolumeManager(base.ManagerWithFind):
         """Delete specified keys from volumes metadata.
 
         :param volume: The :class:`Volume`.
-        :param metadata: A list of keys to be removed.
+        :param keys: A list of keys to be removed.
         """
         for k in keys:
             self._delete("/volumes/%s/metadata/%s" % (base.getid(volume), k))
@@ -377,3 +393,18 @@ class VolumeManager(base.ManagerWithFind):
         return self._action('os-migrate_volume_completion',
                             old_volume,
                             {'new_volume': new_volume_id, 'error': error})[1]
+
+    def update_all_metadata(self, volume, metadata):
+        """Update all metadata of a volume.
+
+        :param volume: The :class:`Volume`.
+        :param metadata: A list of keys to be updated.
+        """
+        body = {'metadata': metadata}
+        return self._update("/volumes/%s/metadata" % base.getid(volume),
+                            body)
+
+    def update_readonly_flag(self, volume, flag):
+        return self._action('os-update_readonly_flag',
+                            base.getid(volume),
+                            {'readonly': flag})
